# Generated by Django 4.0.5 on 2022-07-06 11:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0038_add_starbase_detail"),
    ]

    operations = [
        migrations.AddField(
            model_name="ownercharacter",
            name="error_count",
            field=models.PositiveIntegerField(
                default=0,
                editable=False,
                help_text="Count of ESI errors which happened with this character.",
            ),
        ),
    ]
