{% load i18n %}
{% load structures %}

{% list_title slot_name %}

{% for asset in asset_list %}
    {% list_asset asset.eve_type asset.quantity asset.is_singleton %}
{% empty %}
    <li class="list-group-item text-muted">{% translate "empty" %}</li>
{% endfor %}
