# Generated by Django 2.2.12 on 2020-06-02 11:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0011_last_online"),
    ]

    operations = [
        migrations.AddField(
            model_name="owner",
            name="has_pings_enabled",
            field=models.BooleanField(
                default=True,
                help_text="to enable or disable pinging of notifications for this owner e.g. with @everyone and @here",
            ),
        ),
        migrations.AddField(
            model_name="webhook",
            name="has_pings_enabled",
            field=models.BooleanField(
                default=True,
                help_text="to enable or disable pinging of notifications for this webhook e.g. with @everyone and @here",
            ),
        ),
    ]
